/*
 * $Id: gd32v_hw_i2c.c 500 2021-08-08 19:43:38Z true $
 * begin 20210618 true
 *
 * basically lets us interface with the i2c shit a little easier.
 */

#include "driver/gd32v_hw_i2c.h"



volatile uint8_t i2c_reset;
volatile uint8_t i2c_last_err_reason[2];
volatile uint8_t i2c_state[2];

static dma_parameter_struct i2c_dma_param;
static const uint8_t i2c_dma_tx_ch[2] = {DMA_CH5, DMA_CH3};
//static const uint8_t i2c_dma_rx_ch[2] = {DMA_CH6, DMA_CH4};
static const uint32_t i2c_data_addr[2] = {I2C0_DATA_ADDRESS, I2C1_DATA_ADDRESS};

static uint8_t 	i2c_slave_addr[2];
static uint8_t *i2c_buf[2];
static uint16_t i2c_bytes[2];

static uint8_t i2c_slave_addr[2];

static void (*i2c_cb[2])(uint8_t);

uint32_t i2c_err_cnt[2];


void i2c_config(uint8_t i2c, uint32_t rate, uint8_t slave_addr)
{
	uint32_t periph;

	if (i2c > 1) return;
	periph = i2c ? I2C1 : I2C0;

    // set up data rate
	if (rate) {
		i2c_clock_config(periph, rate, I2C_DTCY_2);
	}

	// set up slave address (ours or theirs)
	i2c_mode_addr_config(periph, I2C_I2CMODE_ENABLE, I2C_ADDFORMAT_7BITS, slave_addr);
	i2c_slave_addr[i2c] = slave_addr << 1;

	// enable the peripheral
    i2c_enable(periph);

	// enable i2c acknowledgement
    i2c_ack_config(periph, I2C_ACK_ENABLE);
}

// initializes i2c hardware.
void i2c_init()
{
	eclic_irq_disable(I2C0_EV_IRQn);
	eclic_irq_disable(I2C0_ER_IRQn);
	eclic_irq_disable(DMA0_Channel5_IRQn);
	dma_deinit(DMA0, i2c_dma_tx_ch[0]);
	i2c_disable(I2C0);

	i2c_interrupt_disable(I2C0, I2C_INT_EV);
	i2c_interrupt_disable(I2C0, I2C_INT_ERR);

	// just set up some defaults
	i2c_dma_param.priority     = DMA_PRIORITY_ULTRA_HIGH;
	i2c_dma_param.memory_inc   = DMA_MEMORY_INCREASE_ENABLE;
	i2c_dma_param.memory_width = DMA_MEMORY_WIDTH_8BIT;
	i2c_dma_param.periph_inc   = DMA_PERIPH_INCREASE_DISABLE;
    i2c_dma_param.periph_width = DMA_PERIPHERAL_WIDTH_8BIT;

	i2c_state[0] = i2c_state[1] = GD32V_I2C_IDLE;

	// enable device clocks
	rcu_periph_clock_enable(RCU_I2C0);

	// enable devices
	i2c_config(0, 400000, 0x80);
	
	eclic_irq_enable(I2C0_EV_IRQn, 12, 0);				// i2c comms interrupt
	eclic_irq_enable(I2C0_ER_IRQn, 0, 0);				// i2c error interrupt
	eclic_irq_enable(DMA0_Channel5_IRQn, 10, 0);		// i2c0 tx
}

void i2c_bus_reset(uint32_t gpio, uint16_t sda, uint16_t scl)
{
	int i;

	gpio_bit_reset(gpio, sda | scl);
	gpio_init(gpio, GPIO_MODE_OUT_PP, GPIO_OSPEED_2MHZ, sda | scl);

	for (i = 0; i < 100; i++) { asm("nop"); }
    gpio_bit_set(gpio, scl);
    for (i = 0; i < 100; i++) { asm("nop"); }
    gpio_bit_set(gpio, sda);
	for (i = 0; i < 100; i++) { asm("nop"); }
	
	gpio_init(gpio, GPIO_MODE_AF_OD, GPIO_OSPEED_2MHZ, sda | scl);
}

void i2c_write(uint8_t i2c, uint8_t slave, const uint8_t *buf, uint16_t bytes, void(*completed_callback)(uint8_t))
{
    uint32_t periph;
	
	if (i2c > 1) return;
	
	// if we're busy, then do nothing
	// todo: handle this better
	if (i2c_state[i2c] != GD32V_I2C_IDLE) return;

	periph = i2c ? I2C1 : I2C0;
	
	// set state
	i2c_state[i2c] = GD32V_I2C_TX_SETUP;
	i2c_cb[i2c] = completed_callback;

	// set configuration
	i2c_config(i2c, 0, slave);
	i2c_buf[i2c] = (uint8_t *)buf;
	i2c_bytes[i2c] = bytes;

	// enable event interrupt then send start on bus; interrupt will take over from here
	i2c_interrupt_enable(periph, I2C_INT_EV);
	i2c_start_on_bus(periph);
}

void i2c_read(uint8_t i2c, uint8_t slave, const char *buf, uint16_t bytes, void(*completed_callback)(uint8_t))
{
	// not yet implemented
}

void i2c_stop(uint8_t i2c)
{
	i2c_stop_on_bus(i2c ? I2C1 : I2C0);
}

void i2c_ev_isr(uint8_t i2c)
{
	uint32_t periph;
	
	periph = i2c ? I2C1 : I2C0;

	// currently hardcoded to be a master transmitter
    if (i2c_interrupt_flag_get(periph, I2C_INT_FLAG_SBSEND)) {
		// start condition recognized, so send slave address
		i2c_state[i2c] = GD32V_I2C_TX_ADDR;
        i2c_master_addressing(periph, i2c_slave_addr[i2c], I2C_TRANSMITTER);
    }

	if (i2c_interrupt_flag_get(periph, I2C_INT_FLAG_ADDSEND)) {
		// disable i2c interrupts. dma will be handling xfer from now
		i2c_interrupt_disable(periph, I2C_INT_EV);
		i2c_interrupt_flag_clear(periph, I2C_INT_FLAG_ADDSEND);

		// address is sent, ready to send data
		i2c_state[i2c] = GD32V_I2C_TX;		

		// set up DMA
		i2c_dma_param.direction   = DMA_MEMORY_TO_PERIPHERAL;
		i2c_dma_param.periph_addr = i2c_data_addr[i2c];
		i2c_dma_param.memory_addr = (uint32_t)i2c_buf[i2c];
		i2c_dma_param.number      = i2c_bytes[i2c];
		dma_init(DMA0, i2c_dma_tx_ch[i2c], &i2c_dma_param);

		// reset and enable DMA interrupt
		dma_interrupt_flag_clear(DMA0, i2c_dma_tx_ch[i2c], DMA_INT_FLAG_G);
		dma_interrupt_enable(DMA0, i2c_dma_tx_ch[i2c], DMA_INT_FTF);
		
		// start DMA transfer
		i2c_dma_enable(periph, I2C_DMA_ON);
		dma_channel_enable(DMA0, i2c_dma_tx_ch[i2c]);
    }

	// if any other interrupts are set, they should not be. clear them.
	// hopefully we've taken care of error condition in error interrupt.
	I2C_STAT0(periph) &= 0xff00;

	/* else

	if (i2c_interrupt_flag_get(I2C0, I2C_INT_FLAG_TBE)) {
        // unused - we're using dma
    }
	*/
}

void i2c_err_isr(uint8_t i2c)
{
	uint32_t periph;
	
	periph = i2c ? I2C1 : I2C0;

	// well fuck
	i2c_dma_enable(periph, I2C_DMA_OFF);

	switch (i2c) {
		case 0: {
			eclic_irq_disable(I2C0_EV_IRQn);
			eclic_irq_disable(DMA0_Channel5_IRQn);
			i2c_bus_reset(GPIOB, GPIO_PIN_7, GPIO_PIN_6);
			break;
		}
	}
	
	i2c_err_cnt[i2c]++;
	i2c_reset |= (1 << i2c);

	// log error
	i2c_last_err_reason[i2c] = I2C_STAT0(periph) >> 8;
	I2C_STAT0(periph) = 0;
	i2c_init();
}

void i2c_dma_complete_cb(uint8_t i2c)
{
	uint32_t periph;
	
	periph = i2c ? I2C1 : I2C0;
	
	// disable i2c interrupts (if for some reason enabled) and send stop on bus
	dma_interrupt_disable(DMA0, i2c_dma_tx_ch[i2c], DMA_INT_FLAG_G);
	i2c_interrupt_disable(periph, I2C_INT_EV);
	i2c_stop_on_bus(periph);
	
	// this transfer is done. execute callback if set
	i2c_state[0] = GD32V_I2C_IDLE;
	if (i2c_cb[i2c]) {
		i2c_cb[i2c](0);
	}
}