/*
 * $Id: w25q_flash.c 499 2021-07-26 05:24:02Z true $
 * begin 20210711 true
 * 
 * winbond w25qxx driver
 */

#include "driver/w25q_flash.h"

uint8_t  cmd[6];
uint32_t old_psc;

static inline void w25q_bspd_fast()
{
	// run at fastest speed
	old_psc = gd32v_spi_get_prescale(W25Q_SPI_DEV);
	gd32v_spi_set_prescale(W25Q_SPI_DEV, SPI_PSC_2);
}

static inline void w25q_bspd_norm()
{
	// reset spi bus back to where it was
	gd32v_spi_set_prescale(W25Q_SPI_DEV, old_psc);
}

static inline void w25q_send(uint8_t bytes)
{
	w25q_bspd_fast();
	
	// do our thing
	W25Q_CS();
	W25Q_SPI_XFER(cmd, cmd, bytes);
	W25Q_NOCS();

	w25q_bspd_norm();
}

void w25q_set_wel(uint8_t enable)
{
	cmd[0] = enable ? W25Q_I_WRITE_ENABLE : W25Q_I_WRITE_DISABLE;
	w25q_send(1);
}

void w25q_power_up()
{
	cmd[0] = W25Q_I_POWER_UP;
	w25q_send(1);

	// wait
	for (int i = 0; i < 200; i++) asm("nop");
}

void w25q_power_down()
{
	cmd[0] = W25Q_I_POWER_DOWN;
	w25q_send(1);
}

int8_t w25q_read(uint32_t addr, uint8_t *dat, uint32_t bytes)
{
	w25q_bspd_fast();

	// set up address then read data
	cmd[0] = W25Q_I_READ;
	cmd[1] = (addr >> 16) & 0xff;
	cmd[2] = (addr >>  8) & 0xff;
	cmd[3] = (addr      ) & 0xff;
	W25Q_CS();
	W25Q_SPI_XFER(0, cmd, 4);           // send read command and address
	W25Q_SPI_XFER(dat, 0, bytes);       // start reading data
	W25Q_NOCS();

	w25q_bspd_norm();

	return 0;
}

int8_t w25q_write(uint32_t addr, uint8_t *dat, uint32_t bytes)
{
	uint16_t cnt;

	if ((bytes > 256) && ((addr & 0xff) != 0x00)) {
		// requesting more than one page written but page doesn't start at page boundary
		// this is not allowed
		return -1;
	}
	
	if (((bytes - 1) & 0xff) + (addr & 0xff) > 0xff) {
		// page crosses program boundary, this is not allowed
		return -2;
	}

	while (bytes) {
		// are we busy? wait for non-busy
		while (w25q_get_sr1() & W25Q_SR1_BUSY);

		// set WEL bit
		w25q_set_wel(1);
		
		// set up address and data length
		cmd[0] = W25Q_I_PAGE_PROGRAM;
		cmd[1] = (addr >> 16) & 0xff;
		cmd[2] = (addr >>  8) & 0xff;
		cmd[3] = (addr      ) & 0xff;

		if (bytes <= 256) {
			cnt = bytes;
			bytes = 0;
		} else {
			cnt = 256;
			bytes -= 256;
		}

		// program this page
		w25q_bspd_fast();

		W25Q_CS();
		W25Q_SPI_XFER(0, cmd, 4);           // send write command and address
		W25Q_SPI_XFER(0, dat, cnt);			// start writing data
		W25Q_NOCS();
		
		// set up address and data pointer for next write
		addr += 256;
		dat += 256;
	}

	w25q_bspd_norm();

	return 0;
}

int8_t w25q_erase_sector(uint32_t addr, uint8_t blocking)
{
	// enable write
	w25q_set_wel(1);
	
	// erase page
	cmd[0] = W25Q_I_SECTOR_ERASE_4K;
	cmd[1] = (addr >> 16) & 0xff;
	cmd[2] = (addr >>  8) & 0xff;
	cmd[3] = (addr      ) & 0xff;
	w25q_send(4);

	if (blocking) {
		while (w25q_get_sr1() & W25Q_SR1_BUSY);
	}

	return 0;
}

uint8_t w25q_get_sr1()
{
	// bits 7 .. 0
	// SRP0 SEC TB BP2 BP1 BP0 WEL BUSY
	cmd[0] = W25Q_I_READ_STATUS_1;
	cmd[1] = 0;
	w25q_send(2);

	return cmd[1];
}

uint8_t w25q_get_sr2()
{
	// bits 7 .. 0
	// x x x x x x QE SRP1
	cmd[0] = W25Q_I_READ_STATUS_2;
	cmd[1] = 0;
	w25q_send(2);

	return cmd[1];
}

uint16_t w25q_get_status()
{
	uint16_t ret;

	ret =  w25q_get_sr1();
	ret |= w25q_get_sr2() << 8;

	return ret;
}

/*
uint8_t w25q_read_uniqueid(uint8_t id[8])
{

}
*/

uint32_t w25q_get_jedec()
{
	cmd[0] = W25Q_I_JEDEC_ID;
	cmd[1] = cmd[2] = cmd[3] = 0;
	w25q_send(4);
	return (cmd[1] << 16) | (cmd[2] << 8) | cmd[3];
}