/*
 * $Id: btn.c 500 2021-08-08 19:43:38Z true $
 * begin 20190527 true
 */

#include "ui/btn.h"

#include "user_config.h"



BtnData btn[BTN_COUNT];

uint8_t btn_data_idx;
uint8_t btn_data[7];

uint8_t btn_read = 0xf;		// buttons idle high
uint32_t btn_invalid_packet_cnt;
uint32_t btn_short_packet_cnt;



void btn_init()
{
	uint8_t i;

	btn_invalid_packet_cnt = 0;
	btn_data_idx = 0;

	for (i = 0; i < BTN_COUNT; i++) {
		btn[i].state = 0;
		btn[i].held = 0;
		btn[i].hold_time = 0;
		btn[i].push_cb = 0;
		btn[i].held_cb = 0;
		btn[i].release_cb = 0;
	}

	// we have a different micro sending button updates via UART.
	// packets are always 4 bytes. so we can set up DMA to get this data,
	// and we can interrupt when we get it to update live state.
	// however our i2c transmitter is using DMA already, so interrupt it is...

	// set up UART for required baud and settings
	rcu_periph_clock_enable(BTN_USART_RCU);

	usart_deinit(BTN_USART);
    usart_baudrate_set(BTN_USART, 100000);
    usart_word_length_set(BTN_USART, USART_WL_8BIT);
    usart_stop_bit_set(BTN_USART, USART_STB_1BIT);
    usart_parity_config(BTN_USART, USART_PM_NONE);
    usart_hardware_flow_rts_config(BTN_USART, USART_RTS_DISABLE);
    usart_hardware_flow_cts_config(BTN_USART, USART_CTS_DISABLE);
    usart_receive_config(BTN_USART, USART_RECEIVE_ENABLE);
    usart_transmit_config(BTN_USART, USART_TRANSMIT_DISABLE);
    usart_enable(BTN_USART);
	
	// configure UART interrupt
	usart_interrupt_flag_clear(BTN_USART, USART_INT_FLAG_RBNE);
	usart_interrupt_enable(BTN_USART, USART_INT_RBNE);
	eclic_irq_enable(BTN_USART_IRQn, BTN_USART_IRQ_LEVEL, 0);
}

void btn_data_process()
{
	int i;
	uint8_t c;
	
	// is data valid?
	c = 0;
	for (i = 0; i < 6; i++) {
		c += btn_data[i];
	}

	// it is valid, or at least likely so
	if (c == btn_data[6]) {
		// use debounce processed data
		btn_read = btn_data[3] & 0x0f;
		// and mark packet as valid
		btn_data[0] = 0x1d;
	} else {
		btn_invalid_packet_cnt++;
	}
}

void btn_data_isr(uint8_t flag)
{	
	if (flag) {
		if (btn_data_idx && btn_data_idx < 7) btn_short_packet_cnt++;
		btn_data_idx = 0;
		usart_interrupt_disable(BTN_USART, USART_INT_IDLE);
	}

	if (btn_data_idx < 7) {
		btn_data[btn_data_idx++] = usart_data_receive(BTN_USART);
		switch (btn_data_idx) {
			case 1: {
				if (btn_data[0] != 0xa5) break;		// invalid header
				usart_interrupt_enable(BTN_USART, USART_INT_IDLE);
				return;
			}
			case 2: {
				if (btn_data[1] != 0x29) break;		// invalid header
				return;
			}
			case 3: {
				if ((btn_data[2] & 0x80) != 0x80) break;
				return;
			}
			case 7: {
				usart_interrupt_disable(BTN_USART, USART_INT_IDLE);
				btn_data_idx = 0;
				btn_data_process();				
				return;
			}
			default: {
				return;
			}
		}
	}

	// we shouldn't normally get here...
	btn_data_idx = 0;
}

static void btn_check(uint8_t idx, uint8_t state)
{
	if (sysflags & SYS_OLED_ROTATE_X) {
		idx += 2;
		idx %= 4;
	}

	if (!state) {	// buttons active low
		if (btn[idx].held < BTN_MAX_HOLD) {
			btn[idx].held++;
		}
	} else {
		btn[idx].held = 0;
	}
}

void btn_tick()
{
	uint8_t i;

	// increment hold
	btn_check(0, btn_read & 4);
	btn_check(1, btn_read & 2);
	btn_check(2, btn_read & 1);
	btn_check(3, btn_read & 8);

	// process
	for (i = 0; i < BTN_COUNT; i++) {
		// test for hold / release, can only happen after push
		if ((btn[i].state & BTN_PUSH) || (btn[i].state & BTN_HELD)) {
			// release first
			if (btn[i].held == 0) {
				// yes, released
				btn[i].state = BTN_RELEASE;

			// do we trigger hold?
			} else if (btn[i].hold_time && (btn[i].held >> BTN_HOLD_SHIFT) == btn[i].hold_time) {
				// yes, held, but only set if we didn't already
				if (!(btn[i].state & BTN_HELD)) {
					btn[i].state = BTN_HELD;
					btn[i].hold_rt_ctr = btn[i].hold_retrig;
				}
			}
		} else {
			// push
			if (btn[i].held == BTN_DEBOUNCE) {
				btn[i].state = BTN_PUSH;
			}
		}
	}
}

void btn_callback()
{
	uint8_t i;
	uint8_t s;

	for (i = 0; i < BTN_COUNT; i++) {
		s = btn[i].state;

		if ((s & BTN_PUSH) && !(s & BTN_PUSH_CB)) {
			btn[i].state |= BTN_PUSH_CB;
			if (btn[i].push_cb) {
				btn[i].push_cb();
			}
		}

		if (btn[i].hold_retrig) {
			if (btn[i].hold_rt_ctr) {
				btn[i].hold_rt_ctr--;
			} else {
				btn[i].hold_rt_ctr = btn[i].hold_retrig;
				btn[i].state &= ~(BTN_HELD_CB);
			}
		}

		if ((s & BTN_HELD) && !(s & BTN_HELD_CB)) {
			btn[i].state |= BTN_HELD_CB;
			if (btn[i].held_cb) {
				btn[i].held_cb();
			}
		}

		if ((s & BTN_RELEASE) && !(s & BTN_RELEASE_CB)) {
			btn[i].state |= BTN_RELEASE_CB;
			if (btn[i].release_cb) {
				btn[i].release_cb();
			}
		}
	}
}
